/*
	KUMAamp project()
	Copyright (C) Mamiya 2000.
	License:GPL
	---------------------------------------------------------------------------
	LZX decoder(interface)
	lzxdecode.c
*/

#include "in_mdx.h"

/* ANSI/Windows standard headers */
#include <windows.h>

/* Project headers */
#include "lzx042.h"

static char tempmdx[MAX_PATH] = "";
static char temppdx[MAX_PATH] = "";

void lzxclean(void)
{
	if (tempmdx[0] != '\0')
	{
		DeleteFile(tempmdx);
		tempmdx[0] = '\0';
	}

	if (temppdx[0] != '\0')
	{
		DeleteFile(temppdx);
		temppdx[0] = '\0';
	}
}

static char *LoadFile(char *fn, LPDWORD lpdwSize)
{
	HANDLE hFile = INVALID_HANDLE_VALUE;
	char *ret = NULL;
	char *buf = NULL;
	do
	{
		DWORD len;
		hFile = CreateFile(
			fn,
			GENERIC_READ,
			FILE_SHARE_READ,
			NULL,
			OPEN_EXISTING,
			FILE_ATTRIBUTE_ARCHIVE |
			FILE_ATTRIBUTE_COMPRESSED |
			FILE_ATTRIBUTE_HIDDEN |
			FILE_ATTRIBUTE_OFFLINE |
			FILE_ATTRIBUTE_READONLY |
			FILE_ATTRIBUTE_SYSTEM |
			FILE_ATTRIBUTE_TEMPORARY |
			FILE_FLAG_SEQUENTIAL_SCAN,
			NULL);
		if (hFile == INVALID_HANDLE_VALUE) break;
		len = GetFileSize(hFile, NULL);
		if (len == 0xFFFFFFFF) break;
		buf = (char *)malloc(len);
		if (buf == NULL) break;
		if (!ReadFile(hFile, buf, len, lpdwSize, NULL)) break;
		if (len != *lpdwSize) break;
		ret = buf;
		buf = NULL;
	} while (0);
	if (hFile != INVALID_HANDLE_VALUE) CloseHandle(hFile);
	if (buf != NULL) free(buf);
	return ret;
}
#define freefile(p) { if (p != NULL) free(p); }

static int GetDwordBE(char *p)
{
	int ret;
	ret  = ((int)(unsigned char)p[0]) << 0x18;
	ret |= ((int)(unsigned char)p[1]) << 0x10;
	ret |= ((int)(unsigned char)p[2]) << 0x08;
	ret |= ((int)(unsigned char)p[3]) << 0x00;
	return ret;
}

static int IsLzxEncoded(char *p)
{
	if (GetDwordBE(p + 4) != 0x4c5a5820) return 0;
	return GetDwordBE(p + 0x12);
}

static int PdxDecode(char *pdxpath, int isLzxMdx)
{
	int ret = 0;
	char *pdxbuf = NULL, *lzxbuf = NULL;
	if (pdxpath[0] == '\0') return ret;
	do
	{
		int i;
		HANDLE hTemp = INVALID_HANDLE_VALUE;
		char temppath[MAX_PATH];
		DWORD pdxsize, lzxsize;
		pdxbuf = LoadFile(pdxpath, &pdxsize);
		if (pdxbuf == NULL)
		{
			lstrcat(pdxpath, ".PDX");
			pdxbuf = LoadFile(pdxpath, &pdxsize);
		}
		if (pdxbuf == NULL) break;
		lzxsize = IsLzxEncoded(pdxbuf);
		if (!isLzxMdx && !lzxsize) break;	/* k */
		if (lzxsize)	/* kPDX */
		{
			lzxbuf = malloc(lzxsize);
			if (lzxbuf == NULL) break;
			lzx042(pdxbuf, lzxbuf);
		}

		GetTempPath(MAX_PATH, temppath);
		for (i = 1; hTemp == INVALID_HANDLE_VALUE; i++)
		{
			wsprintf(temppdx, "%stmp_%d.pdx",temppath, i);
			hTemp = CreateFile(
				temppdx,
				GENERIC_READ | GENERIC_WRITE,
				FILE_SHARE_READ | FILE_SHARE_WRITE/* | FILE_SHARE_DELETE*/,
				NULL,
				CREATE_NEW,
				FILE_ATTRIBUTE_NORMAL/* | FILE_FLAG_DELETE_ON_CLOSE*/,
				NULL
			);
		}
		if (hTemp != INVALID_HANDLE_VALUE)
		{
			if (lzxsize)
				WriteFile(hTemp, lzxbuf, lzxsize, &pdxsize, NULL);
			else
				WriteFile(hTemp, pdxbuf, pdxsize, &lzxsize, NULL);
			CloseHandle(hTemp);
			lstrcpy(pdxpath, temppdx + lstrlen(temppath));
			ret = 1;
		}
		else
		{
			temppdx[0] = '\0';
		}
	} while (0);
	freefile(pdxbuf);
	freefile(lzxbuf);
	return ret;
}

char *lzxdecode(char *infile)
{
	char *mdxbuf = NULL, *lzxbuf = NULL;
	char *ret = infile;
	do
	{
		HANDLE hTemp = INVALID_HANDLE_VALUE;
		char temppath[MAX_PATH];
		char title[MAX_PATH], pdxpath[MAX_PATH], *sp, *dp;
		DWORD mdxsize, lzxsize;
		int i, tlen, plen, datasize;

		sp = mdxbuf = LoadFile(infile, &mdxsize);
		if (mdxbuf == NULL) break;

		for (dp = title; *sp != '\x1a';) *(dp++) = *(sp++);	/* title */
		*(dp++) = '\0';	sp++;
		for (dp = pdxpath; *sp != '\0';) *(dp++) = *(sp++);	/* pdxpath */
		*(dp++) = '\0';	sp++;
		datasize = mdxsize - (sp - mdxbuf);
		lzxsize = IsLzxEncoded(sp);

		if (!PdxDecode(pdxpath, (lzxsize != 0))) break;

		tlen = lstrlen(title);
		plen = lstrlen(pdxpath);

		if (lzxsize != 0) datasize = lzxsize;
		dp = lzxbuf = malloc(tlen + 1 + plen + 1 + datasize);
		if (lzxbuf == NULL) break;

		lstrcpy(dp, title);		dp[tlen] = '\x1a';	dp += tlen + 1; 
		lstrcpy(dp, pdxpath);	dp[plen] = '\0';	dp += plen + 1; 

		if (lzxsize != 0)
			lzx042(sp, dp);
		else
			CopyMemory(dp, sp, datasize);
		dp += datasize;

		GetTempPath(MAX_PATH, temppath);
		for (i = 1; hTemp == INVALID_HANDLE_VALUE; i++)
		{
			wsprintf(temppdx, "%stmp_%d.mdx",temppath, i);
			hTemp = CreateFile(
				tempmdx,
				GENERIC_READ | GENERIC_WRITE,
				FILE_SHARE_READ | FILE_SHARE_WRITE/* | FILE_SHARE_DELETE*/,
				NULL,
				CREATE_NEW,
				FILE_ATTRIBUTE_NORMAL/* | FILE_FLAG_DELETE_ON_CLOSE*/,
				NULL
			);
		}
		if (hTemp != INVALID_HANDLE_VALUE)
		{
			WriteFile(hTemp, lzxbuf, dp - lzxbuf, &mdxsize, NULL);
			CloseHandle(hTemp);
			ret = tempmdx;
		}
		else
		{
			tempmdx[0] = '\0';
		}
	} while(0);

	freefile(mdxbuf);
	freefile(lzxbuf);
	return ret;
}
