/*
	KUMAamp project()
	Copyright (C) Mamiya 2000.
	License:GPL
	---------------------------------------------------------------------------
	MDC to MDX converter
	mdc2mdx.c
*/

#include "in_mdx.h"

/* ANSI/Windows standard headers */
#include <windows.h>

/* Project headers */
#include "lzx042.h"

static char tempmdx[MAX_PATH] = "";

void mdcclean(void)
{
	if (tempmdx[0] != '\0')
	{
		DeleteFile(tempmdx);
		tempmdx[0] = '\0';
	}
}

static char *LoadFile(char *fn, LPDWORD lpdwSize)
{
	HANDLE hFile = INVALID_HANDLE_VALUE;
	char *ret = NULL;
	char *buf = NULL;
	do
	{
		DWORD len;
		hFile = CreateFile(
			fn,
			GENERIC_READ,
			FILE_SHARE_READ,
			NULL,
			OPEN_EXISTING,
			FILE_ATTRIBUTE_ARCHIVE |
			FILE_ATTRIBUTE_COMPRESSED |
			FILE_ATTRIBUTE_HIDDEN |
			FILE_ATTRIBUTE_OFFLINE |
			FILE_ATTRIBUTE_READONLY |
			FILE_ATTRIBUTE_SYSTEM |
			FILE_ATTRIBUTE_TEMPORARY |
			FILE_FLAG_SEQUENTIAL_SCAN,
			NULL);
		if (hFile == INVALID_HANDLE_VALUE) break;
		len = GetFileSize(hFile, NULL);
		if (len == 0xFFFFFFFF) break;
		buf = (char *)malloc(len);
		if (buf == NULL) break;
		if (!ReadFile(hFile, buf, len, lpdwSize, NULL)) break;
		if (len != *lpdwSize) break;
		ret = buf;
		buf = NULL;
	} while (0);
	if (hFile != INVALID_HANDLE_VALUE) CloseHandle(hFile);
	if (buf != NULL) free(buf);
	return ret;
}
#define freefile(p) { if (p != NULL) free(p); }

static int GetDwordBE(char *p)
{
	int ret;
	ret  = ((int)(unsigned char)p[0]) << 0x18;
	ret |= ((int)(unsigned char)p[1]) << 0x10;
	ret |= ((int)(unsigned char)p[2]) << 0x08;
	ret |= ((int)(unsigned char)p[3]) << 0x00;
	return ret;
}
static int GetWordBE(char *p)
{
	int ret;
	ret  = ((int)(unsigned char)p[0]) << 0x08;
	ret |= ((int)(unsigned char)p[1]) << 0x00;
	return ret;
}

#define MDX_BUFSIZE 4096
typedef struct {
	int code;
	int enable;
	int waveform;
	int speed;
	int depth;
	int delay;
	int update;
} lfo;
typedef struct {
	HANDLE hFile;
	char buf[MDX_BUFSIZE];
	int bufp;
	int fofs;
	int pordepth;
	int porstep;
	int tie;
	lfo pm;
	lfo am;
} mdxbuf;
static void initlfo(lfo *plfo)
{
	plfo->enable = 0;
	plfo->waveform = 0;
	plfo->speed = 0;
	plfo->depth = 0;
	plfo->delay = 0;
	plfo->update = 0;
}
static void inittrack(mdxbuf *mdx)
{
	mdx->pordepth = 0;
	mdx->porstep = 0;
	mdx->tie = 0;
	initlfo(&mdx->pm);
	mdx->pm.code = '\xec';
	initlfo(&mdx->am);
	mdx->am.code = '\xeb';
}

static void mdxflush(mdxbuf *mdx)
{
	DWORD len;
	if (mdx->bufp)
	{
		WriteFile(mdx->hFile, mdx->buf, mdx->bufp, &len, NULL);
		mdx->bufp = 0;
	}
}
static void mdxclose(mdxbuf *mdx)
{
	if (mdx != NULL)
	{
		mdxflush(mdx);
		CloseHandle(mdx->hFile);
		free(mdx);
	}
}
static void mdxputc(mdxbuf *mdx, int c)
{
	mdx->fofs++;
	mdx->buf[mdx->bufp++] = c;
	if (mdx->bufp == MDX_BUFSIZE) mdxflush(mdx);
}
static void mdxputw(mdxbuf *mdx, int w)
{
	mdxputc(mdx, w >> 0x08);
	mdxputc(mdx, w >> 0x00);
}

static void mdxwrite(mdxbuf *mdx, char *buf, int l)
{
	if (l)
	{
		if (l <= MDX_BUFSIZE - mdx->bufp)
		{
			CopyMemory(mdx->buf + mdx->bufp, buf, l);
			mdx->bufp += l;
			if (mdx->bufp == MDX_BUFSIZE) mdxflush(mdx);
		}
		else
		{
			DWORD len;
			mdxflush(mdx);
			WriteFile(mdx->hFile, buf, l, &len, NULL);
		}
		mdx->fofs += l;
	}
}

static mdxbuf *mdxopen(void)
{
	int i;
	char temppath[MAX_PATH];
	mdxbuf *mdx = NULL;

	mdx = (mdxbuf *)malloc(sizeof(mdxbuf));
	if (mdx == NULL) return NULL;
	mdx->hFile = INVALID_HANDLE_VALUE;
	mdx->bufp = mdx->fofs = 0;

	GetTempPath(MAX_PATH, temppath);
	for (i = 1; mdx->hFile == INVALID_HANDLE_VALUE; i++)
	{
		wsprintf(tempmdx, "%stmp_%d.mdx",temppath, i);
		mdx->hFile = CreateFile(
			tempmdx,
			GENERIC_READ | GENERIC_WRITE,
			FILE_SHARE_READ | FILE_SHARE_WRITE/* | FILE_SHARE_DELETE*/,
			NULL,
			CREATE_NEW,
			FILE_ATTRIBUTE_NORMAL/* | FILE_FLAG_DELETE_ON_CLOSE*/,
			NULL
		);
	}
	return mdx;
}


char *GetVariable(char *p, int *retp)
{
	int ret = 0;
	do
	{
		ret <<= 7;
		ret |= *p & 0x7f;
	} while (*(p++) & 0x80);
	if (retp != NULL) *retp = ret;
	return p;
}

static char *getloop(char *sp)
{
	while (sp[0] != '\xfe')
	{
		if ('\x00' <= sp[0] && sp[0] <= '\x7f')
		{
			if ('\x00' <= sp[1] && sp[1] <= '\x7f')
			{
				sp = GetVariable(sp + 2, NULL);
				sp = GetVariable(sp, NULL);
			}
			else
			{
				sp += 2;
			}
			continue;
		}
		switch (sp[0])
		{
			case '\xe7':	/* OPMn[hLFOݒ */
				if (sp[1] & 0x80)
				{
					sp = sp + 1 + 1;
				}
				else
				{
					sp = sp + 1 + 5;
				}
				break;
			case '\x80':	/* x */
			case '\xc1':	/* sb`kenXs[hݒ */
			case '\xc4':	/* sb`kenfBCݒ */
			case '\xc6':	/* ʂkenXs[hݒ */
			case '\xc9':	/* ʂkenfBCݒ */
				sp = GetVariable(sp + 1, NULL);
				break;
			case '\x81':	/*  */
				sp = GetVariable(sp + 2, NULL);
				break;
			case '\xb8':	/* |^g */
				sp = GetVariable(sp + 3, NULL);
				break;
			case '\x86':	/* ^C(X[) */
			case '\x8e':	/* ߏI */
			case '\xff':	/* iv[v|Cg}[N */
				sp = sp + 1;
				break;
			case '\xa2':	/* {[ */
			case '\xa3':	/* {[ */
			case '\xa6':	/* p|bgݒ */
			case '\xac':	/* f`[ݒ */
			case '\xae':	/* NI^CYݒ */
			case '\xc0':	/* sb`keng`ݒ */
			case '\xc5':	/* ʂkeng`ݒ */
			case '\xc7':	/* ʂkenUݒ */
			case '\xe0':	/* vO`FW */
			case '\xef':	/* gbNoG[V */
				sp = sp + 1 + 1;
				break;
			case '\x88':	/* s[gJn */
			case '\x89':	/* s[gI[ */
			case '\x8a':	/* s[go */
			case '\xb0':	/* `[ݒ */
			case '\xc2':	/* sb`kenUݒ */
			case '\xe8':	/* OPMWX^ݒ */
			case '\xf0':	/* e|ݒ */
				sp = sp + 1 + 2;
				break;
#if 0
			case '\x8d':	/* Wv */
			case '\x8f':	/* Z[W[ */
				sp = sp + (GetWordBE(sp + 1) ^ 0x8000) - 0x8000;
				break;
#endif
			default:
				return NULL;
		}
	}
	if (sp[1] == '\0' && sp[2] == '\0' ) return NULL;
	return sp + 3 + GetWordBE(sp + 1) - 0x10000;
}

typedef struct {
	int startofs;
	int breakofs;
	lfo pm;
	lfo am;
} loopinfo;
static void mdxlfo(mdxbuf *mdx, lfo *plfo)
{
	if ((plfo->update & 7) == 7 && plfo->speed != 0)
	{
		int clock;
		clock = plfo->depth / plfo->speed;
		if (clock < 0) clock = -clock;
		if (clock == 0) clock = 1;
		mdxputc(mdx, plfo->code);
		mdxputc(mdx, plfo->waveform);
		if (plfo->waveform == 0)
			mdxputw(mdx, clock * 4);
		else
			mdxputw(mdx, clock * 2);
		if (plfo->waveform == 1)
			mdxputw(mdx, plfo->depth / 256);
		else
			mdxputw(mdx, plfo->depth * 256 / clock);
		plfo->update = 8;
		if (!plfo->enable)
		{
			mdxputc(mdx, plfo->code);
			mdxputc(mdx, '\x80');
		}
	}
}
static void rest(mdxbuf *mdx, int step)
{
	mdxlfo(mdx, &mdx->pm);
	mdxlfo(mdx, &mdx->am);
	while (step > 0x80)
	{
		mdxputc(mdx, 0x7f);
		step -= 0x80;
	}
	if (step) mdxputc(mdx, step - 1);
}

static void noteon(mdxbuf *mdx, int note, int vel, int step, int gate)
{
	int porspeed, porstep = 0;
	mdxlfo(mdx, &mdx->pm);
	mdxlfo(mdx, &mdx->am);
	/* O4c 60 -> o4c 0xad */
	note += 0xad-60;
	if (note < 0x80) note = 0x80;
	if (note > 0xdf) note = 0xdf;
	if (vel)
	{
		mdxputc(mdx, '\xfb');
		mdxputc(mdx, vel ^ 0xff);
	}
	if (gate > step) gate = step;
	if (mdx->porstep)
	{
		porspeed = (mdx->pordepth * 256) / mdx->porstep;
		porstep = mdx->porstep;
		if (mdx->tie)
		{
			if (porstep > gate)
			{
				mdx->porstep -= gate;
				mdx->pordepth = (mdx->pordepth * mdx->porstep) / porstep;
				porstep = gate;
			}
			else
			{
				mdx->porstep = 0;
			}
		}
		else
		{
			if (porstep > gate) porstep = gate;
			mdx->porstep = 0;
		}
	}
	step -= gate;
	if (porstep)
	{
		gate -= porstep;
		while (porstep > 0x100)
		{
			mdxputc(mdx, '\xf2');
			mdxputw(mdx, porspeed);
			mdxputc(mdx, note);
			mdxputc(mdx, 0xff);
			mdxputc(mdx, '\xf7');
			porstep -= 0x100;
		}
		mdxputc(mdx, '\xf2');
		mdxputw(mdx, porspeed);
		mdxputc(mdx, note);
		mdxputc(mdx, porstep - 1);
		if (gate) mdxputc(mdx, '\xf7');
	}
	if (gate)
	{
		while (gate > 0x100)
		{
			mdxputc(mdx, note);
			mdxputc(mdx, 0xff);
			mdxputc(mdx, '\xf7');
			gate -= 0x100;
		}
		mdxputc(mdx, note);
		mdxputc(mdx, gate - 1);
	}
	if (step) rest(mdx, step);
	mdx->tie = 0;
}

static void trackconv(mdxbuf *mdx, char *track)
{
	loopinfo loop[32];
	int looplevel = -1;
	int lofs = -1;
	int tmp;
	int note,step,gate,vel;
	char *sp = track, *lp;
	inittrack(mdx);
	lp = getloop(sp);
	while (sp[0] != '\xfe')
	{
		if (lp == sp)
		{
			mdxlfo(mdx, &mdx->pm);
			mdxlfo(mdx, &mdx->am);
			lofs = mdx->fofs;
#if 0
			if (mdx->pm.update & 8) mdx->pm.update = 7;
			mdxlfo(mdx, &mdx->pm);
			if (mdx->am.update & 8) mdx->am.update = 7;
			mdxlfo(mdx, &mdx->am);
#endif
		}
		if ('\x00' <= sp[0] && sp[0] <= '\x7f')
		{
			note = sp[0];
			vel = sp[1];
			if (0 <= vel && vel <= 0x7f)
			{
				sp = GetVariable(sp + 2, &step);
				sp = GetVariable(sp, &gate);
				noteon(mdx, note, vel, step, gate);
			}
			else
			{
				vel = 0;
				step = gate = sp[1] & 0x7f;
				noteon(mdx, note, vel, step, gate);
				sp += 2;
			}
			continue;
		}
		if (sp[0] == '\x80')	/* x */
		{
			sp = GetVariable(sp + 1, &step);
			rest(mdx, step);
			mdx->porstep = 0;
			mdx->tie = 0;
			continue;
		}
		if (sp[0] == '\x81')	/*  */
		{
			note = sp[1];
			sp = GetVariable(sp + 2, &step);
			gate = step;
			vel = 0;
			noteon(mdx, note, vel, step, gate);
			continue;
		}
		switch (sp[0])
		{
			case '\x8e':	/* ߏI */
			case '\xff':	/* iv[v|Cg}[N */
				sp = sp + 1;
				break;
			case '\xef':	/* gbNoG[V */
				sp = sp + 1 + 1;
				break;

			case '\x88':	/* s[gJn */
				mdxlfo(mdx, &mdx->pm);
				mdxlfo(mdx, &mdx->am);
				if (++looplevel > 31)
				{
					mdxputc(mdx, '\xf1');
					mdxputc(mdx, '\x00');
					return;
				}
				loop[looplevel].startofs = mdx->fofs;
				loop[looplevel].breakofs = -1;
				mdxputc(mdx, '\xf6');
				mdxputc(mdx, sp[1]);
				mdxputc(mdx, '\x00');
				sp = sp + 1 + 2;
				break;
			case '\x8a':	/* s[go */
				mdxlfo(mdx, &mdx->pm);
				mdxlfo(mdx, &mdx->am);
				if (looplevel < 0 || loop[looplevel].breakofs != -1)
				{
					mdxputc(mdx, '\xf1');
					mdxputc(mdx, '\x00');
					return;
				}
				CopyMemory(&loop[looplevel].pm, &mdx->pm, sizeof(lfo));
				CopyMemory(&loop[looplevel].am, &mdx->am, sizeof(lfo));
				loop[looplevel].breakofs = mdx->fofs;
				mdxputc(mdx, '\xf4');
				mdxputw(mdx, 0);
				sp = sp + 1 + 2;
				break;
			case '\x89':	/* s[gI[ */
				mdxlfo(mdx, &mdx->pm);
				mdxlfo(mdx, &mdx->am);
				if (looplevel < 0)
				{
					mdxputc(mdx, '\xf1');
					mdxputc(mdx, '\x00');
					return;
				}
				if (loop[looplevel].breakofs != -1)
				{
					char buf[2];
					DWORD l;
					CopyMemory(&mdx->pm, &loop[looplevel].pm, sizeof(lfo));
					CopyMemory(&mdx->am, &loop[looplevel].am, sizeof(lfo));
					mdxflush(mdx);
					SetFilePointer(mdx->hFile, loop[looplevel].breakofs + 1, NULL, FILE_BEGIN);
					tmp = mdx->fofs - loop[looplevel].breakofs - 2;
					buf[0] = (tmp >> 0x08);
					buf[1] = (tmp >> 0x00);
					WriteFile(mdx->hFile, buf, 2, &l, NULL);
					SetFilePointer(mdx->hFile, mdx->fofs, NULL, FILE_BEGIN);
				}
				tmp = loop[looplevel].startofs - mdx->fofs;
				mdxputc(mdx, '\xf5');
				mdxputw(mdx, tmp);
				looplevel--;
				sp = sp + 1 + 2;
				break;

			case '\x86':	/* ^C(X[) */
				mdxputc(mdx, '\xf7');
				mdx->tie = 1;
				sp = sp + 1;
				break;
			case '\xa2':	/* {[ݒ */
				if (sp[1] & 0x80)
				{
					mdxputc(mdx, '\xfb');
					mdxputc(mdx, sp[1] & 0x0f);
				}
				else
				{
					mdxputc(mdx, '\xfb');
					mdxputc(mdx, sp[1] ^ 0xff);
				}
				sp = sp + 1 + 1;
				break;
			case '\xa3':	/* ΃{[ݒ */
				if (sp[1] < 0) {
					for (tmp = 0 ; tmp < -sp[1] ; tmp++) 
						mdxputc(mdx, '\xfa');
				} else {
					for (tmp = 0 ; tmp < sp[1] ; tmp++) 
						mdxputc(mdx, '\xf9');
				}
				sp = sp + 1 + 1;
				break;
			case '\xa6':	/* p|bgݒ */
				tmp = sp[1] - '\xf0';
				if (0 <= tmp && tmp <= 3)
				{
					mdxputc(mdx, '\xfc');
					mdxputc(mdx, tmp);
				}
				sp = sp + 1 + 1;
				break;
			case '\xae':	/* NI^CYݒ */
				mdxputc(mdx, '\xf8');
				if (sp[1])
					mdxputc(mdx, sp[1]);
				else
					mdxputc(mdx, '\x08');
				sp = sp + 1 + 1;
				break;
			case '\xe0':	/* vO`FW */
				mdxputc(mdx, '\xfd');
				mdxputc(mdx, sp[1]);
				sp = sp + 1 + 1;
				break;
			case '\xac':	/* f`[ݒ */
				mdxputc(mdx, '\xf3');
				mdxputw(mdx, sp[1]);
				sp = sp + 1 + 1;
				break;
			case '\xb0':	/* `[ݒ */
				tmp = (GetWordBE(sp + 1) ^ 0x8000) - 0x8000;
				mdxputc(mdx, '\xf3');
				mdxputw(mdx, tmp);
				sp = sp + 1 + 2;
				break;
			case '\xf0':	/* e|ݒ */
				tmp = GetWordBE(sp + 1);
				if (tmp != 0)
					tmp = 256 - (4883 / tmp);
				if (tmp < 0) tmp = 0;
				if (tmp > 255) tmp = 255;
				mdxputc(mdx, '\xff');
				mdxputc(mdx, tmp);
				sp = sp + 1 + 2;
				break;
			case '\xe8':	/* OPMWX^ݒ */
				mdxputc(mdx, '\xfe');
				mdxputc(mdx, sp[1]);
				mdxputc(mdx, sp[2]);
				sp = sp + 1 + 2;
				break;
			case '\xe7':	/* OPMn[hLFOݒ */
				mdxputc(mdx, '\xea');
				if (sp[1] & 0x80)
				{
					mdxputc(mdx, sp[1]);
					sp = sp + 1 + 1;
				}
				else
				{
					mdxwrite(mdx, &sp[1], 5);
					sp = sp + 1 + 5;
				}
				break;

			case '\xb8':	/* |^g */
				mdx->pordepth = (GetWordBE(sp + 1) ^ 0x8000) - 0x8000;
				sp = GetVariable(sp + 3, &mdx->porstep);
				break;

			case '\xc0':	/* sb`keng`ݒ */
				if (sp[1] & 0x80)
				{
					mdxputc(mdx, mdx->pm.code);
					mdxputc(mdx, sp[1]);
					mdx->pm.enable = sp[1] & 1;
					mdxlfo(mdx, &mdx->pm);
				}
				else
				{
					mdx->pm.waveform = sp[1] & 0x7f;
					mdx->pm.update |= 1;
				}
				sp = sp + 1 + 1;
				break;
			case '\xc1':	/* sb`kenXs[hݒ */
				sp = GetVariable(sp + 1, &mdx->pm.speed);
				mdx->pm.update |= 2;
				break;
			case '\xc2':	/* sb`kenUݒ */
				mdx->pm.depth = (GetWordBE(sp + 1) ^ 0x8000) - 0x8000;
				mdx->pm.update |= 4;
				sp = sp + 1 + 2;
				break;
			case '\xc4':	/* sb`kenfBCݒ */
				sp = GetVariable(sp + 1, &mdx->pm.delay);
				tmp = mdx->pm.delay;
				if (tmp > 255) tmp = 255;
				if (tmp < 1)   tmp = 1;
				mdxputc(mdx, '\xe9');
				mdxputc(mdx, tmp);
				break;

			case '\xc5':	/* ʂkeng`ݒ */
				if (sp[1] & 0x80)
				{
					mdxputc(mdx, mdx->am.code);
					mdxputc(mdx, sp[1]);
					mdx->am.enable = sp[1] & 1;
					mdxlfo(mdx, &mdx->am);
				}
				else
				{
					mdx->am.waveform = sp[1] & 0x7f;
					mdx->am.update |= 1;
				}
				sp = sp + 1 + 1;
				break;
			case '\xc6':	/* ʂkenXs[hݒ */
				sp = GetVariable(sp + 1, &mdx->am.speed);
				mdx->am.update |= 2;
				break;
			case '\xc7':	/* ʂkenUݒ */
				mdx->am.depth = sp[1];
				mdx->am.update |= 4;
				sp = sp + 1 + 1;
				break;
			case '\xc9':	/* ʂkenfBCݒ */
				sp = GetVariable(sp + 1, &mdx->am.delay);
				tmp = mdx->am.delay;
				if (tmp > 255) tmp = 255;
				if (tmp < 1)   tmp = 1;
#if 0
				mdxputc(mdx, '\xe9');
				mdxputc(mdx, tmp);
#endif
				break;
#if 0
			case '\x8d':	/* Wv */
			case '\x8f':	/* Z[W[ */
				sp = sp + (GetWordBE(sp + 1) ^ 0x8000) - 0x8000;
				break;
#endif
			default:
				mdxputc(mdx, '\xf1');
				mdxputc(mdx, '\x00');
				return;
		}
	}
	if (lofs != -1)
	{
		mdxlfo(mdx, &mdx->pm);
		mdxlfo(mdx, &mdx->am);
		tmp = lofs - (mdx->fofs + 3);
		mdxputc(mdx, '\xf1');
		mdxputw(mdx, tmp);
	}
	else
	{
		mdxputc(mdx, '\xf1');
		mdxputc(mdx, '\x00');
	}
}
static void toneconv(mdxbuf *mdx, char *tone)
{
	char *sp, *ep;
	int num;
	ep = tone + GetDwordBE(tone);
	num = GetWordBE(tone + 4);
	sp = tone + 6;
	while (num-- && sp < ep)
	{
		mdxputc(mdx, sp[3] + (sp[2] << 7));
		mdxwrite(mdx, sp + 4, 26);
		sp += GetWordBE(sp);
	}
	return;
}

char *mdc2mdx(char *infile)
{
	char *ret = infile;
	char *mdcbuf = NULL;
	mdxbuf *mdx = NULL;
	DWORD mdcsize;
	do
	{
		int i, headofs, ch;
		char head[20];
		char *sp, *seq;
		mdx = mdxopen();
		if (mdx == NULL) break;
		mdcbuf = LoadFile(infile, &mdcsize);
		if (mdcbuf == NULL) break;
		if (mdcsize < 0x40 || GetDwordBE(mdcbuf) != 0x4d44431a) break;
		if (GetDwordBE(mdcbuf + 0x14) != 0)
		{
			sp = mdcbuf + GetDwordBE(mdcbuf + 0x14);
			while (*sp != '\x1a' && *sp != '\0') mdxputc(mdx, *(sp++));
		}
		else
		{
			mdxputc(mdx, '\x0d');
			mdxputc(mdx, '\x0a');
		}
		mdxputc(mdx, '\x1a');
		mdxputc(mdx, '\0');		/* PCM͖Ή */
		headofs = mdx->fofs;	/* ł܂ */
		mdxwrite(mdx, head, 20);	/* _~[ */
		seq = mdcbuf + GetDwordBE(mdcbuf + 0x10);
		for (ch = 0; ch < 8; ch++)
		{
			head[ch * 2 + 2 + 0] = (char)((mdx->fofs - headofs) >> 0x08);
			head[ch * 2 + 2 + 1] = (char)((mdx->fofs - headofs) >> 0x00);
			for (i = 0; i < GetWordBE(seq); i++)
			{
				if (ch == seq[2 + i * 8 + 5]) break;
			}
			if (i == GetWordBE(seq))
			{
				mdxputc(mdx, '\xf1');
				mdxputc(mdx, '\x00');
				continue;
			}
			sp = seq + GetDwordBE(seq + 2 + i * 8);
			trackconv(mdx, sp);
		}
		/* ADPCM`l */
		head[ch * 2 + 2 + 0] = (char)((mdx->fofs - headofs) >> 0x08);
		head[ch * 2 + 2 + 1] = (char)((mdx->fofs - headofs) >> 0x00);
		mdxputc(mdx, '\xf1');
		mdxputc(mdx, '\x00');
		/* F */
		head[0] = (char)((mdx->fofs - headofs) >> 0x08);
		head[1] = (char)((mdx->fofs - headofs) >> 0x00);
		toneconv(mdx, mdcbuf + GetDwordBE(mdcbuf + 0x18));
		/* wb_㏑ */
		mdxflush(mdx);
		SetFilePointer(mdx->hFile, headofs, NULL, FILE_BEGIN);
		WriteFile(mdx->hFile, head, 20, &mdcsize, NULL);
		ret = tempmdx;
	} while(0);
	freefile(mdcbuf);
	mdxclose(mdx);
	if (ret == infile) mdcclean();
	return ret;
}
